{-# LANGUAGE ScopedTypeVariables  #-}
{-# LANGUAGE TypeOperators        #-}
{-# LANGUAGE TypeApplications     #-}
{-# LANGUAGE UndecidableInstances #-}
{-# OPTIONS_GHC -fno-warn-orphans #-}
module Gargantext.Orphans.OpenAPI where

import Control.Lens
import Data.Typeable
import Data.HashMap.Strict.InsOrd qualified as HM
import Data.OpenApi as OpenAPI hiding (Header, Server)
import Data.Swagger qualified as Swagger
import Data.Swagger.Internal.Schema qualified as Swagger
import Data.Text qualified as T
import Prelude
import Servant.API
import Servant.Auth
import Servant.OpenApi

-- Auth instances

instance (HasOpenApi api) => HasOpenApi (Servant.Auth.Auth '[] a :> api) where
  toOpenApi Proxy = toOpenApi $ Proxy @api

instance (HasOpenApi (Servant.Auth.Auth auths a :> api)) => HasOpenApi (Servant.Auth.Auth (Servant.Auth.BasicAuth : auths) a :> api) where
  toOpenApi Proxy = addSecurity $ toOpenApi $ Proxy @(Servant.Auth.Auth auths a :> api)
   where
    addSecurity = addSecurityRequirement identifier . addSecurityScheme identifier securityScheme
    identifier :: T.Text = "BasicAuth"
    securityScheme =
      SecurityScheme
        { _securitySchemeType = SecuritySchemeHttp HttpSchemeBasic
        , _securitySchemeDescription = Just "Basic Authentication"
        }

instance (HasOpenApi (Servant.Auth.Auth auths a :> api)) => HasOpenApi (Servant.Auth.Auth (Servant.Auth.JWT : auths) a :> api) where
  toOpenApi Proxy = addSecurity $ toOpenApi $ Proxy @(Servant.Auth.Auth auths a :> api)
   where
    addSecurity = addSecurityRequirement identifier . addSecurityScheme identifier securityScheme
    identifier :: T.Text = "JWT"
    securityScheme =
      SecurityScheme
        { _securitySchemeType = SecuritySchemeHttp $ HttpSchemeBearer $ Just "JWT"
        , _securitySchemeDescription = Just "Bearer Authentication"
        }

instance (HasOpenApi (Servant.Auth.Auth auths a :> api)) => HasOpenApi (Servant.Auth.Auth (Servant.Auth.Cookie : auths) a :> api) where
  toOpenApi Proxy = addSecurity $ toOpenApi $ Proxy @(Servant.Auth.Auth auths a :> api)
   where
    addSecurity = addSecurityRequirement identifier . addSecurityScheme identifier securityScheme
    identifier :: T.Text = "Cookie"
    securityScheme =
      SecurityScheme
        { _securitySchemeType = SecuritySchemeHttp $ HttpSchemeBearer $ Just "JWT"
        , _securitySchemeDescription = Just "Cookie Authentication"
        }

addSecurityScheme :: T.Text -> SecurityScheme -> OpenApi -> OpenApi
addSecurityScheme securityIdentifier securityScheme openApi =
  openApi
    { _openApiComponents =
        (_openApiComponents openApi)
          { _componentsSecuritySchemes =
              _componentsSecuritySchemes (_openApiComponents openApi)
                <> SecurityDefinitions (HM.singleton securityIdentifier securityScheme)
          }
    }

addSecurityRequirement :: T.Text -> OpenApi -> OpenApi
addSecurityRequirement securityRequirement =
  allOperations
    . security
    %~ ((SecurityRequirement $ HM.singleton securityRequirement []) :)

fromParamSchema :: Swagger.ParamSchema t -> Schema
fromParamSchema ps = OpenAPI.Schema
  { OpenAPI._schemaTitle = Nothing
  , OpenAPI._schemaDescription = Nothing
  , OpenAPI._schemaDefault = Swagger._paramSchemaDefault ps
  , OpenAPI._schemaExample = Nothing
  , OpenAPI._schemaType = convertType <$> Swagger._paramSchemaType ps
  , OpenAPI._schemaFormat = Swagger._paramSchemaFormat ps
  , OpenAPI._schemaNullable = Nothing
  , OpenAPI._schemaEnum = Swagger._paramSchemaEnum ps
  , OpenAPI._schemaMaximum = Swagger._paramSchemaMaximum ps
  , OpenAPI._schemaExclusiveMaximum = Swagger._paramSchemaExclusiveMaximum ps
  , OpenAPI._schemaMinimum = Swagger._paramSchemaMinimum ps
  , OpenAPI._schemaExclusiveMinimum = Swagger._paramSchemaExclusiveMinimum ps
  , OpenAPI._schemaMaxLength = Swagger._paramSchemaMaxLength ps
  , OpenAPI._schemaMinLength = Swagger._paramSchemaMinLength ps
  , OpenAPI._schemaPattern = Swagger._paramSchemaPattern ps
  , OpenAPI._schemaMaxItems = Swagger._paramSchemaMaxItems ps
  , OpenAPI._schemaMinItems = Swagger._paramSchemaMinItems ps
  , OpenAPI._schemaUniqueItems = Swagger._paramSchemaUniqueItems ps
  , OpenAPI._schemaMaxProperties = Nothing
  , OpenAPI._schemaMinProperties = Nothing
  , OpenAPI._schemaAdditionalProperties = Nothing
  , OpenAPI._schemaItems = convertItems <$> Swagger._paramSchemaItems ps
  , OpenAPI._schemaProperties = mempty
  , OpenAPI._schemaRequired = mempty
  , OpenAPI._schemaAllOf = Nothing
  , OpenAPI._schemaOneOf = Nothing
  , OpenAPI._schemaNot = Nothing
  , OpenAPI._schemaAnyOf = Nothing
  , OpenAPI._schemaDiscriminator = Nothing
  , OpenAPI._schemaReadOnly = Nothing
  , OpenAPI._schemaWriteOnly = Nothing
  , OpenAPI._schemaXml = Nothing
  , OpenAPI._schemaExternalDocs = Nothing
  , OpenAPI._schemaDeprecated = Nothing
  , OpenAPI._schemaMultipleOf = Swagger._paramSchemaMultipleOf ps
  }

convertType :: Swagger.SwaggerType t -> OpenAPI.OpenApiType
convertType Swagger.SwaggerString = OpenAPI.OpenApiString
convertType Swagger.SwaggerNumber = OpenAPI.OpenApiNumber
convertType Swagger.SwaggerInteger = OpenAPI.OpenApiInteger
convertType Swagger.SwaggerBoolean = OpenAPI.OpenApiBoolean
convertType Swagger.SwaggerArray = OpenAPI.OpenApiArray
convertType Swagger.SwaggerFile = OpenAPI.OpenApiString -- OpenAPI lacks file type, fallback to string
convertType Swagger.SwaggerNull = OpenAPI.OpenApiNull
convertType Swagger.SwaggerObject = OpenAPI.OpenApiObject

convertItems :: Swagger.SwaggerItems t -> OpenAPI.OpenApiItems
convertItems (Swagger.SwaggerItemsPrimitive _collectionFormat paramSchema) =
  OpenAPI.OpenApiItemsObject $ OpenAPI.Inline $ fromParamSchema paramSchema
convertItems (Swagger.SwaggerItemsObject referencedSchema) =
  OpenAPI.OpenApiItemsObject $ convertRef referencedSchema
convertItems (Swagger.SwaggerItemsArray referencedSchemas) =
  OpenAPI.OpenApiItemsArray $ map convertRef referencedSchemas

convertRef :: Swagger.Referenced Swagger.Schema -> OpenAPI.Referenced OpenAPI.Schema
convertRef (Swagger.Ref ref) = OpenAPI.Ref $ OpenAPI.Reference (Swagger.getReference ref)
convertRef (Swagger.Inline s) = OpenAPI.Inline $ fromSwaggerSchema s

convertAdditionalProps :: Swagger.AdditionalProperties -> OpenAPI.AdditionalProperties
convertAdditionalProps (Swagger.AdditionalPropertiesAllowed allowed) =
  OpenAPI.AdditionalPropertiesAllowed allowed
convertAdditionalProps (Swagger.AdditionalPropertiesSchema ref) =
  OpenAPI.AdditionalPropertiesSchema (convertRef ref)

convertDocs :: Swagger.ExternalDocs -> OpenAPI.ExternalDocs
convertDocs Swagger.ExternalDocs { Swagger._externalDocsDescription = desc
                                 , Swagger._externalDocsUrl = docurl } =
  OpenAPI.ExternalDocs
    { OpenAPI._externalDocsDescription = desc
    , OpenAPI._externalDocsUrl = convertURL docurl
    }

convertURL :: Swagger.URL -> OpenAPI.URL
convertURL = OpenAPI.URL . Swagger.getUrl

convertXml :: Swagger.Xml -> OpenAPI.Xml
convertXml Swagger.Xml { Swagger._xmlName = xname
                       , Swagger._xmlNamespace = xnamespace
                       , Swagger._xmlPrefix = xprefix
                       , Swagger._xmlAttribute = xattribute
                       , Swagger._xmlWrapped = xwrapped } =
  OpenAPI.Xml
    { OpenAPI._xmlName = xname
    , OpenAPI._xmlNamespace = xnamespace
    , OpenAPI._xmlPrefix = xprefix
    , OpenAPI._xmlAttribute = xattribute
    , OpenAPI._xmlWrapped = xwrapped
    }

convertDiscriminator :: T.Text -> OpenAPI.Discriminator
convertDiscriminator t = OpenAPI.Discriminator t mempty

fromSwaggerSchema :: Swagger.Schema -> OpenAPI.Schema
fromSwaggerSchema swaggerSchema = OpenAPI.Schema
  { OpenAPI._schemaTitle = Swagger._schemaTitle swaggerSchema
  , OpenAPI._schemaDescription = Swagger._schemaDescription swaggerSchema
  , OpenAPI._schemaDefault = Nothing
  , OpenAPI._schemaExample = Swagger._schemaExample swaggerSchema
  , OpenAPI._schemaType = Nothing
  , OpenAPI._schemaFormat = Nothing
  , OpenAPI._schemaNullable = Nothing
  , OpenAPI._schemaEnum = Nothing
  , OpenAPI._schemaMaximum = Nothing
  , OpenAPI._schemaExclusiveMaximum = Nothing
  , OpenAPI._schemaMinimum = Nothing
  , OpenAPI._schemaExclusiveMinimum = Nothing
  , OpenAPI._schemaMaxLength = Nothing
  , OpenAPI._schemaMinLength = Nothing
  , OpenAPI._schemaPattern = Nothing
  , OpenAPI._schemaMaxItems = Nothing
  , OpenAPI._schemaMinItems = Nothing
  , OpenAPI._schemaUniqueItems = Nothing
  , OpenAPI._schemaMaxProperties = Swagger._schemaMaxProperties swaggerSchema
  , OpenAPI._schemaMinProperties = Swagger._schemaMinProperties swaggerSchema
  , OpenAPI._schemaAdditionalProperties = convertAdditionalProps <$> Swagger._schemaAdditionalProperties swaggerSchema
  , OpenAPI._schemaItems = Nothing
  , OpenAPI._schemaProperties = HM.map convertRef $ Swagger._schemaProperties swaggerSchema
  , OpenAPI._schemaRequired = Swagger._schemaRequired swaggerSchema
  , OpenAPI._schemaAllOf = map convertRef <$> Swagger._schemaAllOf swaggerSchema
  , OpenAPI._schemaOneOf = Nothing
  , OpenAPI._schemaNot = Nothing
  , OpenAPI._schemaAnyOf = Nothing
  , OpenAPI._schemaDiscriminator = convertDiscriminator <$> Swagger._schemaDiscriminator swaggerSchema
  , OpenAPI._schemaReadOnly = Swagger._schemaReadOnly swaggerSchema
  , OpenAPI._schemaWriteOnly = Nothing
  , OpenAPI._schemaXml = convertXml <$> Swagger._schemaXml swaggerSchema
  , OpenAPI._schemaExternalDocs = convertDocs <$> Swagger._schemaExternalDocs swaggerSchema
  , OpenAPI._schemaDeprecated = Nothing
  , OpenAPI._schemaMultipleOf = Nothing
  }

-- little compat-shim to avoid re-implementing 'ToSchema' and 'ToParamSchema' for each
-- and every type, while we transition away from swagger.

instance {-# OVERLAPPABLE #-} Swagger.ToParamSchema a => ToParamSchema a where
  toParamSchema p = fromParamSchema $ Swagger.toParamSchema p

instance {-# OVERLAPPABLE #-} (Typeable a, Swagger.ToSchema a) => ToSchema a where
  declareNamedSchema _ = pure $ OpenAPI.NamedSchema swaggerName openApiSchema
    where
      Swagger.NamedSchema swaggerName swaggerSchema = Swagger.toNamedSchema (Proxy :: Proxy a)
      openApiSchema = fromSwaggerSchema swaggerSchema
